#!/Users/jacky/tools/venv/bin/python3.9
import json
import copy
from functools import total_ordering
from collections import OrderedDict
from termcolor import colored


class TableNode(object):
    
    def __init__(self,*args,**kwargs) -> None:
        super().__init__()
        for k,v in kwargs.items():
            setattr(self,k,v)
        
        
    def __len__(self):
        return len(self.__dict__)
    
    # return True if every key and value is same and with same length
    def __eq__(self, another: object) -> bool:
        if len(self) != len(another): return False
        for p in self.__dict__: # check every property's value is same
            if self.__dict__.get(p) != another.__dict__.get(p): return False 
        return True     
    
    # Return True if self and another has top three values in same, which we regard they are some record
    # The purpose of doing so is to check if two records are actually same, so has to be merged.
    def is_same_record(self, another: object):
        obj=copy.deepcopy(self)
        try:
            top_three=list(obj.__dict__)[0:3]
        except:
            raise IndexError(f'{self} has less than 3 columns')
        for property in top_three:
            if getattr(obj,property) !=getattr(another,property): 
                return False
        return True 
    
    # + operator： if another has value, no matter self has value or not,  replace self’s value with another’s value, but if another has no value, keep self’s value. 
    def __add__(self,another:object):
        obj=copy.deepcopy(self)
        # 相同变量部分：搜索another每一个值，如果不为空，就替代self。 
        # 不同变量部分，another多出来的部分,增加变量和值到self。
        for p in another.__dict__:
            if p in self.__dict__:
                # If another has value, no matter self has value or not,  replace self’s value with another’s value
                if getattr(another,p,None): 
                    setattr(obj,p,getattr(another,p))
                # but if another has no value, keep self’s value. 
            else:
                # if a property in another but not in self, add it to self 
                setattr(obj,p,getattr(another,p,None))
                
        return obj
        
    # copy:  if another has value, no matter self has value or not,  replace self’s value with another’s value
    def copy(self,another:object):
        obj=copy.deepcopy(self)
        # 相同变量部分：搜索another每一个值，替代self。 
        for p in another.__dict__:
            if p in self.__dict__ : setattr(obj,p,getattr(another,p))
        return obj
        
    def __hash__(self) -> int:
        hs=[]
        for key, value in self.__dict__.items():
            hs.append('{} = {}'.format(key, value))
        return hash(','.join(hs))
    
    # return the variable name representing the object
    def __str__(self):
        str_str=['{}:{}'.format(key, value) for key, value in self.__dict__.items()]
        return ', '.join(str_str)
    
    # return the string representing the class constructor method
    def __repr__(self):
        repr_str=['{} = "{}"'.format(key, value)  if type(value)==str else '{} = {}'.format(key, value) for key, value in self.__dict__.items()]
        return f'TableNode({",".join(repr_str)})'
    
    # return True or False representing node is null or not null. 
    # null means the infoNode has no any value, reflecting in excel is a blank row
    def __bool__(self):
        return any(self.__dict__.values())

# sheet data as a dict xx 
@total_ordering
class TableList(object):
    """[TableList class used for handle inside  of table sheet and between of table sheets level information]

    Args:
        2D list including three  rows of sheet title list, column title list, and variables list
        columns are defined by excel files. 

    Raises:
        StopIteration: [description]
        KeyError: ['description']

    Returns:
        TableList object
    """    
    def __init__(self,rows) -> None:
        super().__init__()
        self.rows=rows
        self.sheet_title=self._getSheetTitle() # the first line is the sheet's title
        self.column_titles=self.rows.pop(0) # the second line is the display title
        # self.column_variables=[self._title2variable(x) for x in variables]# the third line in rows is this sheet's schema, used as title variables
        self.column_variables=self._cleanTitles()
        self.column_titles=OrderedDict(zip(self.column_variables,self.column_titles)) #绑定title到variable，否则在后续+ - table的时候，title和variable 以及value可能无法保持一致。 
        self.rows=rows  # data rows
        self.data=[]  # contains the sheet's all TableNode objects
        self.createTableNodes(rows) #rows after poped three rows
        self.index=0
    
    def _getSheetTitle(self):
        st=self.rows.pop(0)
        st=st[0].split('\n')[0] if len(st)>0 else "A TableList object"
        return st
    
    def _cleanTitles(self):
        raw_variables=self.rows.pop(0)
        variables=[x for x in raw_variables if x!=None]
        if len(variables)<len(raw_variables): print(colored(f"There are some blank columns inside the table sheet {self.sheet_title}, you may delete those blank columns at the right of your visiable columns",'red'))
        return variables
    
    # TODO: convert?
    def _title2variable(self,title):
        return title.lower().replace(' ','_')
    #TODO: ?
    def _variable2title(self,variable):
        return variable.replace('_',' ').title()
    
    def __str__(self):
        return self.sheet_title
    
    # hash only care about variables and values
    def __hash__(self) -> int:
        return hash(','.join([node for node in self.data]))
    
    def __len__(self):
        return len(self.data)
    
    #TODO: has same value. __eq__ only cares about has same schema, instead of values. So, has same value will compare value
    def has_same_value(self,another):
        pass
    
    # == > < All this operations do not consider value. They only compare sheet variables
    # == return True if self and another has same variables. Only care about variables
    def __eq__(self, another: object) -> bool:
        # check if sheet has same column variables, which means same data structure
        if another==None: return False
        if self.data==[] and another==[]: return True
        if self.column_variables!=another.column_variables: return False 
        return True 
    
    # # > return True if every another object's items is in self, and the length of self is greater than the other's
    def __gt__(self,another:object)->bool:
        flags=[]
        for variable in another.column_variables:
            flags.append(True) if variable in self.column_variables else flags.append(False)
        flags.append(len(self.column_variables)>len(another.column_variables))
        return all(flags)  
    
    # + union two objects (self | another)'s two lists(column_titles, column_variables), and return a new object
    def __add__(self,another:object):
        obj=copy.deepcopy(self) #deep copy self to obj, to avoid messing up self object
        # get titles united. list the OrderedDict union result for remove duplication 
        # get variables
        obj.column_variables=list(OrderedDict.fromkeys(self.column_variables+another.column_variables)) #字典keys相加去重复
        titles=[self.column_titles.get(v) or another.column_titles.get(v) for v in obj.column_variables] # get new set of titles
        obj.column_titles=OrderedDict(zip(obj.column_variables,titles)) # re-bind to new set of variables
        another.column_titles,another.column_variables= obj.column_titles, obj.column_variables #让another和obj一样的titles和variables
        
        # expand obj and another to the new union. As a result, they will be in same number of columns
        for o in obj.data:
            for p in obj.column_variables:
                if not getattr(o,p,None): # if the property has no value, set the property as None.
                    setattr(o,p,None)
                    
        for o in another.data:
            for p in another.column_variables: # here obj.column_variables are expaned.
                if not getattr(o,p,None):
                    setattr(o,p,None)

        # Now self and another are in same dimension in columns。 
        temp_data=[]
        for node_obj in obj.data: # 对每一个obj里面的数据 去寻找是否和another里面的是相同记录
            for node_another in another.data:
                if node_obj.is_same_record(node_another): # if same record
                    node_obj+=another.data.pop(another.data.index(node_another)) # pop up the same one and merge
                    temp_data.append(node_obj) #将merge过的记录添加到临时list中
                else:   
                    if node_obj not in temp_data:   #如果不是同一条记录，而且在临时列表中没有，添加，
                        temp_data.append(node_obj) 

        obj.data=temp_data+another.data # append all different rows to obj.data。 
        #此事obj。data数据都有了，但是order有可能发生改变，所以下面需要对obj的data进行重构，以variables的顺序作为统一标准。        
        # 根据obj.data中node的variable排序，重新构建obj。
        new_data=[]
        for d in obj.data:
            values=[getattr(d,p,None) for p in obj.column_variables] #以variable取出对应的value
            pair=OrderedDict(zip(obj.column_variables,values))  #zip variable和value，编程有序字典
            new_data.append(TableNode(**pair)) #重新构建TableNode obj到临时data列表
        obj.data=new_data       
        return obj
    
    # - operator: 1st: minus only allowed if self and another have same column schema(same variables). 2st, if top 3 variables are same, the row will be removed
    def __sub__(self,another:object):
        obj=copy.deepcopy(self) #deep copy self to obj, to avoid messing up self object
        # check if has same column schema, if not raise an error
        if self.column_variables!=another.column_variables:
            raise ValueError(f"{self.column_titles} is not same as {another.column_variables}, so could not minus")
        # check for same row, if yes, remove it
        for obj_node in self.data: # use self.data because obj.data is changing...
            for another_node in another.data:
                if obj_node.is_same_record(another_node):
                    obj.data.pop(obj.data.index(obj_node))
        
        return obj
    
    # in: return True if self's variables contains another's
    def __contain__(self,another: object):
        for x in another.column_variables:
            if x not in self.column_variables: return False
        return True 
    
    def createTableNodes(self,rows):
        for row in rows:
            row_data=OrderedDict(zip(self.column_variables,row)) # zip and make data as a dict
            self.data.append(TableNode(**row_data)) # create a InfoNode object
            
    def showTableNodes(self,seperator=','):
        for row in self.data:
            info=[info for info in row.__dict__.values()]
            info=[ str(i) if i is not None else '' for i in info] #avoid None
            print(seperator.join(info))
    
    
    def copy(self,another: object):
        obj=copy.deepcopy(self)
        # for some tables, has to be specially processed
        if another.name in ["personid","address","phone","eraddress","contact"]:
            # loop another.data
            for ad in another.data:
                # loop obj.data
                for index,od in enumerate(obj.data): 
                    # if another.data's variable ==obj.data's variable_type, then copy
                    if ad.variable_type==od.variable_type and ad.display_type==od.display_type: # using display_type is only for working_address, which has multiple value in er.xlsx
                        obj.data[index]=od.copy(ad)
                        break
            return obj
        #for general tables
        for ad in another.data:
            # if has same schema
            if self.column_variables==another.column_variables:
                # if another not in original one, just append another's rows to original one.
                if ad not in obj.data: obj.data.append(ad)
            # if has different schema
            else:
                # get self's column variable,and get value from another, then make a TableNode object and insert to self.data
                kwargs={}
                for key in obj.column_variables:
                    kwargs[key]=getattr(ad,key,None)
                copied_obj=TableNode(**kwargs)
                obj.data.append(copied_obj)
        return obj
    
    #get common variables in self and another, and return new object
    def common(self,another: object):
        obj=copy.deepcopy(self)
        # get titles united. list the OrderedDict union result for remove duplication 
        obj.column_titles=[{k:v} for k,v in self.column_titles.items() if k in another.column_titles.keys()]
        # get variables
        obj.column_variables=[x for x in self.column_variables if x in another.column_variables]
        #get values
        obj.data=[x for x in self.data if x in another.data]
        return obj
        
    # Iterator
    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self.data[self.index]
        except IndexError:
            raise StopIteration()
        self.index += 1
        return item
        



#!/Users/jacky/tools/venv/bin/python3.9
# coding=utf-8
# from pydantic import ValidationError
import json
from termcolor import colored
# from openpyxl import load_workbook, Workbook
import copy
from functools import total_ordering
from collections import OrderedDict


# info sheet data structure , in which 'variable','tag','description','value' are MUST.
class InfoNode(object):
    
    def __init__(self,*args,**kwargs) -> None:
        super().__init__()
        for k,v in kwargs.items():
            setattr(self,k,v)
    
    def __len__(self):
        return len(self.__dict__)
    
    #TODO: has same value. __eq__ only cares about has same schema, instead of values. So, has same value will compare value
    # def has_same_value(self,another):
    #     pass
    
    # Only care about variable and its value, if both same then True 
    def __eq__(self, another: object) -> bool:
        if another==None: 
            return False
        return True if self.variable==another.variable and self.value==another.value else False
    
    # + operator： if another has value, no matter self has value or not,  replace self’s value with another’s value, but if another has no value, keep self’s value. 
    def __add__(self,another:object):
        obj=copy.deepcopy(self)
        if obj.variable!=another.variable:
            raise TypeError(f'The two nodes ({obj.variable} {another.variable}) is different, can not add up')
        if another.value:
            obj.value=another.value
        return obj

    # - operator： if a node obj in self and in another, and if another has value, no matter self has value or not,  delete self’s node obj, but if another has no value, keep self’s node obj. 
    def __sub__(self,another:object):
        obj=copy.deepcopy(self)
        if obj.variable!=another.variable:
            raise TypeError(f'The two nodes ({obj.variable} {another.variable}) is different, can not minus ')
        if another.value:
            return None
        return obj
        
    def __hash__(self) -> int:
        return hash(self.variable+str(self.value))
    
    # return the variable name representing the object
    def __str__(self):
        return self.variable
    
    # return the string representing the class constructor method
    def __repr__(self):
        repr_str=[]
        for key, value in self.__dict__.items():
            repr_str.append('{} = "{}"'.format(key, value) ) if type(value)==str else repr_str.append('{} = {}'.format(key, value) )
        return f'InfoNode({",".join(repr_str)})'
    
# sheet data as a dict xx 
@total_ordering
class SheetDict(object):
    """[SheetDict class used for handle inside  of sheet and between of sheets level information]

    Args:
        2D list including three  rows of sheet title list, column title list, and variables list
        The data rows must include 'variable','tag','description', and 'value' four columns. 
        Other columns are optional. 

    Raises:
        StopIteration: [description]
        KeyError: ['description']

    Returns:
        SheetDict object
    """
    
    def __init__(self,rows) -> None:
        super().__init__()
        self.rows=rows
        self.sheet_title=self._getSheetTitle() # the first line is the sheet's title
        self.column_titles=self.rows.pop(0) # the second line is the display title
        variables=self._cleanTitles()
        # self.column_variables=[x and str(x).lower() for x in variables]# the third line in rows is this sheet's schema, used as title variables TODO: 不要小写？ 
        self.column_variables=[x and str(x) for x in variables]
        self.data=OrderedDict()  # contains the sheet's all variable and InfoNode objects ordered dict
        self._createInfoNodes(rows) #rows after poped three rows
        self.index=0
    
    def _getSheetTitle(self):
        st=self.rows.pop(0)
        st=st[0].split('\n')[0] if len(st)>0 else "A SheetDict object"
        return st
        
    def _cleanTitles(self):
        raw_variables=self.rows.pop(0)
        variables=[x for x in raw_variables if x!=None]
        if len(variables)<len(raw_variables): print(colored(f"There are some blank columns inside the info sheet {self.sheet_title}, you may delete those blank columns at the right of your visiable columns",'red'))
        return variables
        
    def __str__(self):
        return self.sheet_title
    
    # hash only care about variables and values
    def __hash__(self) -> int:
        return hash(','.join(['{} = {}'.format(key, value) for key, value in self.data.items()]))
    
    def __len__(self):
        return len(self.data)
    
    #TODO: has same value. __eq__ only cares about has same schema, instead of values. So, has same value will compare value
    def has_same_value(self,another):
        pass
    
    # define sheet data's collection methods, + -  == > < All this operations do not consider value. They only compare sheet variables
    # == return True if self and another has same variables. Only care about variables
    def __eq__(self, another: object) -> bool:
        if another==None: 
            return False
        if self is None:
            return False
        if self.data.keys()!=another.data.keys(): return False 
        # for k, v in self.data.items():
        #     if v.value != another.data[k].value: return False
        return True 
    
    # > return True if every another object's items is in self, and the length of self is greater than the other's
    def __gt__(self,another:object)->bool:
        flags=[]
        for variable in another.data:
            flags.append(True) if variable in self.data else flags.append(False)
        flags.append(len(self)>len(another))
        return all(flags)  
    
    # + union two sets (self | another), and return a new set
    def __add__(self,another:object):
        obj=copy.deepcopy(self) #deep copy self to obj, to avoid messing up self object
        # iterate another and insert or replace self's：
        for variable in another.data:
            if variable not in self.data:   #if node is not in self, add another’s node in self’s. 
                obj.data[another.data[variable].variable]=another.data[variable]
            else:   #if node is in self, add up two nodes. 
                obj.data[variable]+=another.data[variable]
            
        return obj
    
    # - set difference, return a new set with elements in self but not in another
    def __sub__(self,another:object):
        obj=copy.deepcopy(self) #deep copy self to obj, to avoid messing up self object
        for variable in self.data:
            # self and another are in same sheet name，sub the variables in self if it's in the another 
            if variable in another.data and another.data[variable].value:
                obj.data.pop(variable,None)
                # obj.data[variable]-=another.data[variable]
            # obj.data=[obj.data[x] for x in obj.data if x!=None]
        return obj
    
    # in: return True if self contains another, else False
    def __contains__(self,another:object):
        # if every another object's items is in self, then return True;
        flags=[]
        for variable in another.data:
            flags.append(True) if variable in self.data else flags.append(False)
        return all(flags)
    
    # Iterator
    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = list(self.data.values())[self.index]
        except IndexError:
            raise StopIteration()
        self.index += 1
        return item
    
    def _createInfoNodes(self,rows):
        if len(rows)==0:
            self.data={}
        for row in rows:
            row_data=OrderedDict(zip(self.column_variables,row)) # zip and make data as a dict
            v=row_data.get('variable')  # get the variable as key
            self.data[v]=InfoNode(**row_data) # create a InfoNode object
            
    def showInfoNodes(self,seperator=','):
        for variable in self.data.values():
            info=[info for info in variable.__dict__.values()]
            info=[ i if i is not None else '' for i in info] #avoid None
            print(seperator.join(info))
    
    # This method returns all info variables or specific info notes specified in program    
    def getInfoNodes(self,program=None):
        if program:
            if program not in self.column_variables:
                raise KeyError(f'{program} is not a valid column name')
            # if a progam marked any word first letter is Y/y, it will be True, otherwise, it will be False
            return {k:v for k,v in self.data.items() if  getattr(v,program)!= None and getattr(v,program)[0].upper()=='Y'}
        return {k:v for k,v in self.data.items()}

    # get common variables in self and another, and return a new object
    def common(self,another: object):
        obj=copy.deepcopy(self) #deep copy self to obj, to avoid messing up self object
        # get two dicts' intersection set
        common_variables=set(self.data.keys()) & set(another.data.keys())
        for variable in self.data:
            if variable not in common_variables: 
                obj.data.pop(variable,None)
        return obj
    
    # copy: if self and another has common variables, then copy the value of another's shared variable to self 
    def copy(self,another):
        obj=copy.deepcopy(self)
        for k,v in another.data.items():
            if k in obj.data.keys(): obj.data[k]=v
        return obj







    """_summary_

    Returns:
        _type_: _description_
    """

class Area:
    """ Data infrastructure for NOC wage and outlook """
    data = [
        ["Canada", "Canada", 0],
        ["Newfoundland and Labrador", "Newfoundland and Labrador", 1],
        ["Newfoundland and Labrador", "Avalon Peninsula Region", 2],
        ["Newfoundland and Labrador", "Notre Dame-Central-Bonavista Bay Region", 3],
        ["Newfoundland and Labrador", "South Coast–Burin Peninsula Region", 4],
        ["Newfoundland and Labrador", "West Coast–Northern Peninsula–Labrador Region", 5],
        ["Prince Edward Island", "Prince Edward Island", 6],
        ["Nova Scotia", "Nova Scotia", 7],
        ["Nova Scotia", "Annapolis Valley Region", 8],
        ["Nova Scotia", "Cape Breton Region", 9],
        ["Nova Scotia", "Halifax Region", 10],
        ["Nova Scotia", "North Shore Region", 11],
        ["Nova Scotia", "Southern Region", 12],
        ["New Brunswick", "New Brunswick", 13],
        ["New Brunswick", "Campbellton–Miramichi Region", 14],
        ["New Brunswick", "Edmundston–Woodstock Region", 15],
        ["New Brunswick", "Fredericton–Oromocto", 16],
        ["New Brunswick", "Moncton–Richibucto Region", 17],
        ["New Brunswick", "Saint John–St. Stephen Region", 18],
        ["Quebec", "Quebec", 19],
        ["Quebec", "Abitibi-Témiscamingue Region", 20],
        ["Quebec", "Bas-Saint-Laurent Region", 21],
        ["Quebec", "Capitale-Nationale Region", 22],
        ["Quebec", "Centre-du-Québec Region", 23],
        ["Quebec", "Chaudière-Appalaches Region", 24],
        ["Quebec", "Côte-Nord Region", 25],
        ["Quebec", "Estrie Region", 26],
        ["Quebec", "Gaspésie–Îles-de-la-Madeleine Region", 27],
        ["Quebec", "Lanaudière Region", 28],
        ["Quebec", "Laurentides Region", 29],
        ["Quebec", "Laval Region", 30],
        ["Quebec", "Mauricie Region", 31],
        ["Quebec", "Montréal Region", 32],
        ["Quebec", "Montérégie Region", 33],
        ["Quebec", "Nord-du-Québec Region", 34],
        ["Quebec", "Outaouais Region", 35],
        ["Quebec", "Saguenay–Lac-Saint-Jean Region", 36],
        ["Ontario", "Ontario", 37],
        ["Ontario", "Hamilton–Niagara Peninsula Region", 38],
        ["Ontario", "Kingston–Pembroke Region", 39],
        ["Ontario", "Kitchener–Waterloo–Barrie Region", 40],
        ["Ontario", "London Region", 41],
        ["Ontario", "Muskoka–Kawarthas Region", 42],
        ["Ontario", "Northeast Region", 43],
        ["Ontario", "Northwest Region", 44],
        ["Ontario", "Ottawa Region", 45],
        ["Ontario", "Stratford–Bruce Peninsula Region", 46],
        ["Ontario", "Toronto Region", 47],
        ["Ontario", "Windsor-Sarnia Region", 48],
        ["Manitoba", "Manitoba", 49],
        ["Manitoba", "Interlake Region", 50],
        ["Manitoba", "North Central Region", 51],
        ["Manitoba", "North Region", 52],
        ["Manitoba", "Parklands Region", 53],
        ["Manitoba", "South Central Region", 54],
        ["Manitoba", "Southeast Region", 55],
        ["Manitoba", "Southwest Region", 56],
        ["Manitoba", "Winnipeg Region", 57],
        ["Saskatchewan", "Saskatchewan", 58],
        ["Saskatchewan", "Northern Region", 59],
        ["Saskatchewan", "Prince Albert Region", 60],
        ["Saskatchewan", "Regina–Moose Mountain Region", 61],
        ["Saskatchewan", "Saskatoon–Biggar Region", 62],
        ["Saskatchewan", "Swift Current–Moose Jaw Region", 63],
        ["Saskatchewan", "Yorkton–Melville Region", 64],
        ["Alberta", "Alberta", 65],
        ["Alberta", "Athabasca–Grande Prairie–Peace River Region", 66],
        ["Alberta", "Banff–Jasper–Rocky Mountain House Region", 67],
        ["Alberta", "Calgary Region", 68],
        ["Alberta", "Camrose–Drumheller Region", 69],
        ["Alberta", "Edmonton Region", 70],
        ["Alberta", "Lethbridge–Medicine Hat Region", 71],
        ["Alberta", "Red Deer Region", 72],
        ["Alberta", "Wood Buffalo–Cold Lake Region", 73],
        ["British Columbia", "British Columbia", 74],
        ["British Columbia", "Cariboo Region", 75],
        ["British Columbia", "Kootenay Region", 76],
        ["British Columbia", "Lower Mainland–Southwest Region", 77],
        ["British Columbia", "Nechako Region", 78],
        ["British Columbia", "North Coast Region", 79],
        ["British Columbia", "Northeast Region", 80],
        ["British Columbia", "Thompson–Okanagan Region", 81],
        ["British Columbia", "Vancouver Island and Coast Region", 82],
        ["Yukon Territory", "Yukon Territory", 83],
        ["Northwest Territories", "Northwest Territories", 84],
        ["Nunavut", "Nunavut", 85]
        ]
    
    #col 1 is index for prince in wage/outlook search, col 2,3 is area range, 4 is index of province for provincial median wage search
    prov_area=[     
        [1,2,5,4],    # Newfoundland and Labrador
        [6,6,6,9],    # Prince Edward Island
        [7,8,12,6],   # Nova Scotia
        [13,14,18,3], # New Brunswick
        [19,20,36,10], # Quebec
        [37,38,48,8], # Ontario
        [49,50,57,2], # Manitoba
        [58,59,64,11], # Saskatchewan
        [65,66,73,0], # Alberta
        [74,75,82,1], # British Columbia
        [83,83,83,12], # Yukon Territory
        [84,84,84,5], # Northwest Territories
        [85,85,85,7]  # Nunavut
    ]

    def __init__(self,areaId=77):
        self.areaId=areaId

    def index(self,area_name):
        index=0
        for area in Area.data:
            if area_name==area[1]:
                return index
            index+=1

    # for wage/outlook
    @property
    def prov_index(self):  
        for pa in Area.prov_area:
            if (self.areaId>=pa[1]) & (self.areaId<=pa[2]):
                return pa[0] 

    # for provincial median wage https://www.canada.ca/en/employment-social-development/services/foreign-workers/service-tables.html
    @property
    def prov_index_median_wage(self):  
        for pa in Area.prov_area:
            if (int(self.areaId)>=pa[1]) & (int(self.areaId)<=pa[2]):
                return pa[3] 
    @property
    def prov_name(self):
        for ai in Area.data:
            if ai[2]==int(self.areaId):
                return ai[0]
    @property
    def area_name(self):
        for ai in Area.data:
            if ai[2]==int(self.areaId):
                return ai[1]


from webform.models.bcpnpmodel_app import BcpnpEEModelApp, BcpnpModelApp
from webform.models.definition import Action
from webform.models.bcpnpmodel_reg import BcpnpModelReg, BcpnpEEModelReg
from webform.bcpnp.dashboard import DashboardApp, DashboardReg
from model.common.educationbase import EducationHistory
from webform.bcpnp.data import (
    getEducation,
    country_map,
    province_map,
    city_map,
    getFieldofStudy,
)
from typing import Union


class EducationApp:
    def __init__(self, person: Union[BcpnpModelApp, BcpnpEEModelApp]):
        self.person = person

    @property
    def high_school(self):
        high_school_value = []
        high_school = EducationHistory(self.person.education).high_school

        for index, edu in enumerate(high_school):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#BCPNP_App_EduSec_From-" + str(index),
                    "value": edu.start_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#BCPNP_App_EduSec_To-" + str(index),
                    "value": edu.end_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Checkbox.value,
                    "label": "Currently Enrolled",
                    "id": f"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(2) > div > uf-panel-body > div > div.ng-pristine.ng-untouched.ng-valid.ng-scope.ng-isolate-scope.ng-not-empty > div:nth-child({index+1}) > uf-row:nth-child(2) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div > label",
                    "value": False if edu.graduate_date else True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of Institution",
                    "id": "#BCPNP_App_EduSec_School-" + str(index),
                    "value": edu.school_name,
                    "length": 200,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City/town",
                    "id": "#BCPNP_App_EduSec_City-" + str(index),
                    "value": edu.city,
                    "length": 100,
                    "required": True,
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Country",
                    "id": "#BCPNP_App_EduSec_Country-" + str(index),
                    "value": country_map[edu.country],
                },
                {
                    "action_type": Action.Radio.value,
                    "label": "Did you successfully complete high school",
                    "id": "#BCPNP_App_EduSec_Completion-" + str(index) + "-Yes"
                    if edu.graduate_date
                    else "#BCPNP_App_EduSec_Completion-" + str(index) + "-No",
                },
            ]
            high_school_value.append(block)
        return [
            {
                "action_type": Action.RepeatSection.value,
                "button_id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(2) > div > uf-panel-body > div > div:nth-child(2) > uf-clone-repeatable > a > i",
                "value": high_school_value,
            }
        ]

    @property
    def post_cecondary_in_bc(self):
        value = []
        education = EducationHistory(self.person.education).post_secondary_in_ca_prov(
            "BC"
        )

        has_post_edu_in_bc = {
            "action_type": Action.Radio.value,
            "label": "Have you enrolled in or completed a post-secondary program within B.C.?",
            "id": "#BCPNP_App_EduBC-Yes"
            if len(education) > 0
            else "#BCPNP_App_EduBC-No",
        }
        for index, edu in enumerate(education):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#BCPNP_App_EduBC_From-" + str(index),
                    "value": edu.start_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#syncA_App_EduBC_To-" + str(index),
                    "value": edu.end_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Checkbox.value,
                    "label": "Currently Enrolled",
                    "id": f"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(3) > div > uf-panel-body > div > div > div.ng-pristine.ng-untouched.ng-valid.ng-isolate-scope.ng-not-empty > div:nth-child({index+1}) > uf-row:nth-child(1) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div > label",
                    "value": False if edu.graduate_date else True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of Institution",
                    "id": "#syncA_App_EduBC_Institution-" + str(index),
                    "value": edu.school_name,
                    "length": 200,
                    "required": True,
                },
                {
                    "action_type": Action.Select.value,
                    "label": "City/town",
                    "id": "#syncA_App_EduBC_City-" + str(index),
                    "value": city_map[edu.city],
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Level of post-secondary education attained",
                    "id": "#syncA_App_EduBC_Level-" + str(index),
                    "value": getEducation(
                        edu.education_level, edu.is_trade, is_reg=False
                    ),
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Field of study",
                    "id": "#syncA_App_EduBC_Field-" + str(index),
                    "value": getFieldofStudy(edu.field_of_study),
                },
            ]
            other_field = (
                [
                    {
                        "action_type": Action.Input.value,
                        "label": "If other, enter field of study",
                        "id": "#BCPNP_App_EduBC_OtherField-" + str(index),
                        "value": edu.field_of_study,
                        "length": 200,
                        "required": True,
                    }
                ]
                if getFieldofStudy(edu.field_of_study)
                == "OTH"  # "Other" returned value is OTH
                else []
            )

            value.append(block + other_field)
        return [
            has_post_edu_in_bc,
            {
                "action_type": Action.RepeatSection.value,
                "button_id": "body > div.page > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(3) > div > uf-panel-body > div > div > div:nth-child(2) > uf-clone-repeatable > a > i",
                "value": value,
            },
        ]

    @property
    def post_secondary_in_canada(self):
        value = []
        education = EducationHistory(
            self.person.education
        ).post_secondary_in_ca_but_not_in_prov("BC")

        has_post_edu_in_ca = {
            "action_type": Action.Radio.value,
            "label": "Have you enrolled in or completed a post-secondary program in Canada but not in B.C.?",
            "id": "#BCPNP_App_EduCAN-Yes"
            if len(education) > 0
            else "#BCPNP_App_EduCAN-No",
        }
        for index, edu in enumerate(education):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#BCPNP_App_EduCAN_From-" + str(index),
                    "value": edu.start_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#syncA_App_EduCAN_To-" + str(index),
                    "value": edu.end_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Checkbox.value,
                    "label": "Currently Enrolled",
                    "id": f"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(4) > div > uf-panel-body > div > div > div.ng-pristine.ng-untouched.ng-valid.ng-isolate-scope.ng-not-empty > div:nth-child({index+1}) > uf-row:nth-child(1) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div > label",
                    "value": False if edu.graduate_date else True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of Institution",
                    "id": "#syncA_App_EduCAN_Institution-" + str(index),
                    "value": edu.school_name,
                    "length": 200,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City/town",
                    "id": "#BCPNP_App_EduCAN_City-" + str(index),
                    "value": edu.city,
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Province/territory",
                    "id": "#BCPNP_App_EduCAN_Province-" + str(index),
                    "value": province_map[edu.province],
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Level of post-secondary education attained",
                    "id": "#syncA_App_EduCAN_Level-" + str(index),
                    "value": getEducation(
                        edu.education_level, edu.is_trade, is_reg=False
                    ),
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Field of study",
                    "id": "#syncA_App_EduCAN_Field-" + str(index),
                    "value": getFieldofStudy(edu.field_of_study),
                },
            ]
            other_field = (
                [
                    {
                        "action_type": Action.Input.value,
                        "label": "If other, enter field of study",
                        "id": "#BCPNP_App_EduCAN_OtherField-" + str(index),
                        "value": edu.field_of_study,
                        "length": 200,
                        "required": True,
                    }
                ]
                if getFieldofStudy(edu.field_of_study)
                == "OTH"  # "Other" returned value is OTH
                else []
            )

            value.append(block + other_field)
        return [
            has_post_edu_in_ca,
            {
                "action_type": Action.RepeatSection.value,
                "button_id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(4) > div > uf-panel-body > div > div > div:nth-child(2) > uf-clone-repeatable > a > i",
                "value": value,
            },
        ]

    @property
    def post_secondary_out_of_canada(self):
        value = []
        education = EducationHistory(self.person.education).post_secondary_not_in_ca

        has_post_edu_out_of_ca = {
            "action_type": Action.Radio.value,
            "label": "Have you enrolled in or completed a post-secondary program outside of Canada?",
            "id": "#BCPNP_App_EduNCAN-Yes"
            if len(education) > 0
            else "#BCPNP_App_EduNCAN-No",
        }
        for index, edu in enumerate(education):
            block = [
                {
                    "action_type": Action.Input.value,
                    "label": "From",
                    "id": "#BCPNP_App_EduPostSec_From-" + str(index),
                    "value": edu.start_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "To",
                    "id": "#BCPNP_App_EduPostSec_To-" + str(index),
                    "value": edu.end_date,
                    "length": 10,
                    "required": True,
                },
                {
                    "action_type": Action.Checkbox.value,
                    "label": "Currently Enrolled",
                    "id": f"body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(5) > div > uf-panel-body > div > div > div.ng-pristine.ng-untouched.ng-valid.ng-isolate-scope.ng-not-empty > div:nth-child({index+1}) > uf-row:nth-child(1) > div > uf-date:nth-child(2) > div > div.form-inline.present-checkbox-container > div > label",  # TODO: see if needed to change
                    "value": False if edu.graduate_date else True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "Name of Institution",
                    "id": "#BCPNP_App_EduPostSec_Institution-" + str(index),
                    "value": edu.school_name,
                    "length": 200,
                    "required": True,
                },
                {
                    "action_type": Action.Input.value,
                    "label": "City/town",
                    "id": "#BCPNP_App_EduPostSec_City-" + str(index),
                    "value": edu.city,
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Country",
                    "id": "#BCPNP_App_EduPostSec_Country-" + str(index),
                    "value": country_map[edu.country],
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Level of post-secondary education attained",
                    "id": "#BCPNP_App_EduPostSec_Level-" + str(index),
                    "value": getEducation(
                        edu.education_level, edu.is_trade, is_reg=False
                    ),
                },
                {
                    "action_type": Action.Select.value,
                    "label": "Field of study",
                    "id": "#BCPNP_App_EduPostSec_Field-" + str(index),
                    "value": getFieldofStudy(edu.field_of_study),
                },
            ]
            other_field = (
                [
                    {
                        "action_type": Action.Input.value,
                        "label": "If other, enter field of study",
                        "id": "#BCPNP_App_EduPostSec_OtherField-" + str(index),
                        "value": edu.field_of_study,
                        "length": 200,
                        "required": True,
                    }
                ]
                if getFieldofStudy(edu.field_of_study)
                == "OTH"  # "Other" returned value is OTH
                else []
            )

            value.append(block + other_field)
        return [
            has_post_edu_out_of_ca,
            {
                "action_type": Action.RepeatSection.value,
                "button_id": "body > div.page > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > div.tab-content > ng-transclude > uf-tab:nth-child(2) > div > uf-panel:nth-child(5) > div > uf-panel-body > div > div > div:nth-child(2) > uf-clone-repeatable > a > i",
                "value": value,
            },
        ]

    # def fill(self):
    #     dashboard = DashboardApp()
    #     actions = (
    #         dashboard.jump("Education")
    #         + self.high_school
    #         + self.post_cecondary_in_bc
    #         + self.post_secondary_in_canada
    #         + self.post_secondary_out_of_canada
    #         + dashboard.save
    #     )
    #     return [
    #         {
    #             "action_type": Action.WebPage.value,
    #             "page_name": "Education",
    #             "actions": actions,
    #             "id": None,
    #         }
    #     ]

    def fill(self):
        # dashboard = DashboardApp()
        actions = (
            # dashboard.jump("Education")
            self.high_school
            + self.post_cecondary_in_bc
            + self.post_secondary_in_canada
            + self.post_secondary_out_of_canada
            # + dashboard.save
        )
        return [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Education",
                "actions": actions,
                "id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li:nth-child(3) > a",
                "next_page_tag":"#BCPNP_App_WorkExp-No"
            }
        ]



from pydantic import BaseModel,validator,EmailStr
from datetime import date
from typing import Optional
from model.common.utils import makeList
from model.common.person import Person
from model.common.mixins import DurationMixin
from model.common.employmentbase import EmploymentBase
from model.common.educationbase import EducationBase

class PersonalAssess(BaseModel):
    self_description:Optional[str]
    skill_list:list
    activity:Optional[list]
    
    _str2bool_activity=validator('activity',allow_reuse=True,pre=True)(makeList)
    _str2bool_skill_list=validator('skill_list',allow_reuse=True,pre=True)(makeList)

class Personal(Person):
    email:Optional[EmailStr]
    
    @property
    def birth_day(self):
        return self.dob.strftime('%b %d, %Y')
    
class Education(EducationBase):
    city:str
    country:str
    description:Optional[str]

class Language(BaseModel):
    reading:str
    writting:str
    listening:str
    speaking:str
    test_type: str
    remark: str
    
class Employment(EmploymentBase):
    duties:list
    
    _str2bool_duties=validator('duties',allow_reuse=True,pre=True)(makeList)




from model.common.commonmodel import CommonModel
from datetime import date
from pydantic import BaseModel
from typing import List

class Travel(BaseModel):
   start_date:date	
   end_date:date	
   length:int	
   destination:str
   purpose:str

class Personal(BaseModel):
    last_name:str
    first_name:str
    
class M5562Model(CommonModel):
    travel:List[Travel]
    personal:Personal
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            excels=self.getExcels(['excel/pa.xlsx'])
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        super().__init__(excels,output_excel_file,globals())



from datetime import date
from turtle import pos

# definition format variables
yes_no = {"validate": "list", "source": ["Yes", "No"]}
bcpnp_case_stream = {
    "validate": "list",
    "source": (
        [
            "EE-Skilled Worker",
            "EE-International Graduate",
            "EE-International Post-Graduate",
            "EE-Health Authority",
            "Skilled Worker",
            "International Graduate",
            "Entry-Level and Semi-Skilled Worker",
            "International Post-Graduate",
            "Health Authority",
        ]
    ),
}
corporate_structure = {
    "validate": "list",
    "source": [
        "Incorporated",
        "Limited Liability Partnership",
        "Extra-provincially-registered",
        "federally-incorporated",
        "Other",
    ],
}
canada_provinces = {
    "validate": "list",
    "source": [
        "AB",
        "BC",
        "MB",
        "NB",
        "NL",
        "NS",
        "NT",
        "NU",
        "ON",
        "PE",
        "QC",
        "SK",
        "YT",
    ],
}
english_french = {
    "validate": "list",
    "source": ["English", "French", "Both", "Neither"],
}
english_or_french = {"validate": "list", "source": ["English", "French"]}
english_french_chinese = {
    "validate": "list",
    "source": ["Chinese", "English", "French"],
}
imm_status = {
    "validate": "list",
    "source": [
        "Citizen",
        "Permanent Resident",
        "Worker",
        "Student",
        "Visitor",
        "Refugess",
        "Other",
    ],
}

salary_payment_way = {
    "validate": "list",
    "source": ["weekly", "bi-weekly", "semi-monthly", "monthly"],
}
wage_unit = {"validate": "list", "source": ["hourly", "weekly", "monthly", "annually"]}
lmia_duration_unit = {"validate": "list", "source": ["month", "year"]}
ot_after_hours_unit = {"validate": "list", "source": ["day", "week"]}
job_duration_unit = {"validate": "list", "source": ["day", "week", "month", "year"]}
purpose_of_lmia = {
    "validate": "list",
    "source": [
        "Support Permanent Resident only",
        "Support Work Permit only",
        "Support both Work Permit and Permanent Resident",
    ],
}
stream_of_lmia = {
    "validate": "list",
    "source": [
        "1. EE(Express Entry: 5593)",
        "2. HWS(High Wage Stream: 5626)",
        "3. LWS(Low Wage Stream: 5627)",
        "4. GTS(Global Talent Stream: 5624, 5625)",
        "5. AC (Academic: 5626)",
        "6. AG (Agriculture: 5519, 5510)",
    ],
}
rent_unit = {"validate": "list", "source": ["week", "month"]}
accommodation_type = {
    "validate": "list",
    "source": ["house", "apartment", "dorm", "other"],
}
sex = {"validate": "list", "source": ["Male", "Female"]}
workpermit_type = {
    "validate": "list",
    "source": [
        "Co-op Work Permit",
        "Exemption from Labour Market Impact Assessment",
        "Labour Market Impact Assessment Stream",
        "Live-in Caregiver Program",
        "Open Work Permit",
        "Open work permit for vulnerable workers",
        "Other",
        "Post Graduation Work Permit",
        "Start-up Business Class",
    ],
}
marital_status = {
    "validate": "list",
    "source": [
        "Annulled Marriage",
        "Common-Law",
        "Divorced",
        "Married",
        "Separated",
        "Single",
        "Unknown",
        "Widowed",
    ],
}
pre_relationship_type = {"validate": "list", "source": ["Common-Law", "Married"]}
language_test_type = {"validate": "list", "source": ["IELTS", "CELPIP", "TEF", "TCF"]}
education_level = {
    "validate": "list",
    "source": [
        "Doctor",
        "Master",
        "Post-graduate diploma",
        "Bachelor",
        "Associate",
        "Diploma/Certificate",
        "High school",
        "Less than high school",
    ],
}
relationship = {
    "validate": "list",
    "source": [
        "Grand Parent",
        "Parent",
        "Spouse",
        "Child",
        "Grand Child",
        "Sibling",
        "Aunt",
        "Uncle",
        "Niece",
        "Newphew",
        "Friend",
    ],
}
family_relationship = {
    "validate": "list",
    "source": ["Spouse", "Son", "Daughter", "Mother", "Father", "Brother", "Sister"],
}
pr_imm_program = {"validate": "list", "source": ["Economic", "Family"]}
pr_imm_category = {
    "validate": "list",
    "source": [
        "Provincial Nominee Program (PNP)",
        "Atlantic Immigration Program",
        "Spouse",
        "Common-law Partner",
        "Dependent Child",
        "Other Relative",
    ],
}
pr_imm_under = {
    "validate": "list",
    "source": [
        "Spouse or common-law partner in Canada class",
        "Family class (outside Canada)",
    ],
}
interview_canadian_status = {
    "validate": "list",
    "source": ["Citizen", "PR", "Foreigner", "Unknown"],
}
tr_application_purpose = {
    "validate": "list",
    "source": ["apply or extend", "restore status", "TRP"],
}
sp_paid_person = {"validate": "list", "source": ["Myself", "Parents", "Other"]}
vr_application_purpose = {
    "validate": "list",
    "source": ["apply or extend visitor record", "restore status as visotor", "TRP"],
}
sp_in_application_purpose = {
    "validate": "list",
    "source": ["apply or extend study permit", "restore status as student", "TRP"],
}
sp_apply_wp_type = {
    "validate": "list",
    "source": ["Co-op Work Permit", "Open Work Permit", "Post Graduation Work Permit"],
}
wp_in_application_purpose = {
    "validate": "list",
    "source": [
        "apply WP for same employer",
        "apply WP for new employer",
        "restore status as worker",
        "TRP with same employer",
        "TRP with new employer",
    ],
}
wp_apply_wp_type = {
    "validate": "list",
    "source": [
        "Co-op Work Permit",
        "Exemption from Labour Market Impact Assessment",
        "Labour Market Impact Assessment Stream",
        "Live-in Caregiver Program",
        "Open Work Permit",
        "Open Work Permit for Vulnerable Workers",
        "Other",
        "Post Graduation Work Permit",
        "Start-up Business Class",
    ],
}
visa_application_purpose = {"validate": "list", "source": ["Visitor Visa", "Transit"]}
tr_original_purpose = {
    "validate": "list",
    "source": ["Business", "Tourism", "Study", "Work", "Other", "Family Visit"],
}

date_format = {
    "validate": "date",
    "criteria": ">",
    "value": date(1800, 1, 1),
    "input_title": "请输入日期",
    "input_message": "格式:1999-06-30",
}
positive_int = {
    "validate": "integer",
    "criteria": ">",
    "value": 0,
    "input_title": "请输入整数:",
    "input_message": "必须是大于0",
}

positive_int_decimal = {
    "validate": "decimal",
    "criteria": ">",
    "value": 0,
    "input_title": "请输入整数或小数:",
    "input_message": "必须是大于0",
}

int_decimal = {
    "validate": "decimal",
    "criteria": "",
    # "value": 0,
    "input_title": "请输入整数或小数:",
    "input_message": "无论正数负数",
}

eye_color = {
    "validate": "list",
    "source": ["Black", "Brown", "Blue", "Green"],
}

employment_type = {
    "validate": "list",
    "source": ["Employed", "Self-employed"],
}


# Validation data
validation = {
    # BCPNP
    "info-bcpnp": {
        "has_applied_before": yes_no,
        "case_stream": bcpnp_case_stream,
        "submission_date": date_format,
        "q1": yes_no,
        "q2": yes_no,
        "q3": yes_no,
        "q4": yes_no,
        "q5": yes_no,
        "q6": yes_no,
        "q7": yes_no,
    },
    # Employer
    "info-general": {
        "ft_employee_number": positive_int,
        "pt_pt_employee_number": positive_int,
        "establish_date": date_format,
        "has_lmia_approved": yes_no,
        "corporate_structure": corporate_structure,
        "num_pnps": positive_int,
        "num_pnps_approved": positive_int,
        "num_pnps_in_process": positive_int,
        "when_lmia_approved": date_format,
        "last_revenue": positive_int_decimal,
        "last_profit": positive_int_decimal,
        "retained_earning": positive_int_decimal,
        "before_last_revenue": positive_int_decimal,
        "before_last_profit": positive_int_decimal,
        "has_jobbank_account": yes_no,
        "has_bc_employer_certificate": yes_no,
    },
    "table-eraddress": {"province": canada_provinces},
    "table-contact": {
        "preferred_language": english_french,
        "province": canada_provinces,
    },
    "table-finance": {
        "year": positive_int,
    },
    "table-employee_list": {
        "wage": positive_int_decimal,
        "hours_per_week": positive_int_decimal,
        "employment_start_date": date_format,
        "immigration_status": imm_status,
    },
    "info-position": {
        "worked_working": yes_no,
        "under_cba": yes_no,
        "has_same": yes_no,
        "lowest": positive_int_decimal,
        "highest": positive_int_decimal,
        "lmia_refused": yes_no,
    },
    "info-joboffer": {
        "license_request": yes_no,
        "union": yes_no,
        "hours": positive_int_decimal,
        "payment_way": salary_payment_way,
        "wage_unit": wage_unit,
        "wage_rate": positive_int_decimal,
        "ot_ratio": positive_int_decimal,
        "ot_after_hours": positive_int_decimal,
        "ot_after_hours_unit": ot_after_hours_unit,
        "is_working": yes_no,
        "work_start_date": date_format,
        "permanent": yes_no,
        "job_duration": positive_int_decimal,
        "job_duration_unit": job_duration_unit,
        "has_probation": yes_no,
        "probation_duration": positive_int_decimal,
        "disability_insurance": yes_no,
        "dental_insurance": yes_no,
        "empolyer_provided_persion": yes_no,
        "extended_medical_insurance": yes_no,
        "english_french": yes_no,
        "other_language_required": yes_no,
        "oral": positive_int_decimal,
        "writing": positive_int_decimal,
        "offer_date": date_format,
        "vacation_pay_weeks": positive_int_decimal,
        "vacation_pay_percentage": positive_int_decimal,
    },
    # LMIA
    "table-employee_list": {"immigration_status": imm_status},
    "info-lmiacase": {
        "area_index": positive_int,
        "province_index": positive_int,
        "unemploy_rate": positive_int_decimal,
        "area_median_wage": positive_int_decimal,
        "noc_outlook": positive_int,
        "provincial_median_wage": positive_int_decimal,
        "is_in_10_days_priority": yes_no,
        "top10_wages": positive_int_decimal,
        "is_waived_from_advertisement": yes_no,
        "purpose_of_lmia": purpose_of_lmia,
        "stream_of_lmia": stream_of_lmia,
        "has_another_employer": yes_no,
        "number_of_tfw": positive_int,
        "duration_number": positive_int_decimal,
        "duration_unit": lmia_duration_unit,
        "has_attestation": yes_no,
    },
    "info-lmi": {
        "laid_off_in_12": yes_no,
        "is_work_sharing": yes_no,
        "labour_dispute": yes_no,
        "canadian_lost_job": yes_no,
        "laid_off_canadians": positive_int,
        "laid_off_tfw": positive_int,
    },
    "info-emp5624": {
        "hird_canadian": yes_no,
        "why_not": yes_no,
        "has_active_lmbp": yes_no,
    },
    "table-emp5624lmbp": {"start_date": date_format, "end_date": date_format},
    "info-emp5626": {
        "is_seasonal": yes_no,
        "start_month": positive_int,
        "end_month": positive_int,
        "last_canadian_number": positive_int,
        "last_tfw_number": positive_int,
        "current_canadian_number": positive_int,
        "current_tfw_number": positive_int,
        "tp_waivable": yes_no,
        "has_finished_tp": yes_no,
    },
    "info-emp5627": {
        "provide_accommodation": yes_no,
        "rent_unit": rent_unit,
        "accommodation_type": accommodation_type,
        "rent_amount": positive_int,
        "bedrooms": positive_int,
        "people": positive_int,
        "bathrooms": positive_int,
    },
    # PA
    "info-personal": {
        "sex": sex,
        "height": positive_int,
        "eye_color": eye_color,
        "dob": date_format,
        "english_french": english_french,
        "which_one_better": english_or_french,
        "net_asset": positive_int,
        "liquid_asset": positive_int,
        "language_test": yes_no,
        "did_eca": yes_no,
        "intended_province": canada_provinces,
        "primary_school_years": positive_int,
        "secondary_school_years": positive_int,
        "post_secondary_school_years": positive_int,
        "other_school_years": positive_int,
    },
    "table-personid": {"issue_date": date_format, "expiry_date": date_format},
    "info-status": {
        "current_country_status": imm_status,
        "current_status_type": workpermit_type,
        "has_vr": yes_no,
        "current_status_start_date": date_format,
        "current_status_end_date": date_format,
        "last_entry_date": date_format,
        "last_entry_place": date_format,
    },
    "info-ee": {"ee_expiry_date": date_format, "ee_score": positive_int},
    "info-marriage": {
        "marital_status": marital_status,
        "married_date": date_format,
        "sp_in_canada": yes_no,
        "sp_language_type": language_test_type,
        "sp_language_r": positive_int_decimal,
        "sp_language_w": positive_int_decimal,
        "sp_language_l": positive_int_decimal,
        "sp_language_s": positive_int_decimal,
        "sp_canada_status": imm_status,
        "previous_married": yes_no,
        "pre_relationship_type": pre_relationship_type,
        "pre_sp_dob": date_format,
        "pre_start_date": date_format,
        "pre_end_date": date_format,
    },
    "table-assumption": {
        "hourly_rate": positive_int_decimal,
        "start_date": date_format,
        "end_date": date_format,
        "work_permit_type": workpermit_type,
        "province": canada_provinces,
    },
    "table-phone": {"number": positive_int},
    "table-history": {"start_date": date_format, "end_date": date_format},
    "table-addresshistory": {"start_date": date_format, "end_date": date_format},
    "table-language": {
        "test_type": language_test_type,
        "date": date_format,
        "report_date": date_format,
        "reading": positive_int_decimal,
        "writting": positive_int_decimal,
        "listening": positive_int_decimal,
        "speaking": positive_int_decimal,
    },
    "table-education": {
        "start_date": date_format,
        "end_date": date_format,
        "education_level": education_level,
        "is_trade": yes_no,
        "academic_year": positive_int_decimal,
        "graduate_date": date_format,
    },
    "table-employment": {
        "start_date": date_format,
        "end_date": date_format,
        "weekly_hours": positive_int_decimal,
        "employment_type": employment_type,
        "share_percentage": positive_int_decimal,
        "bcpnp_qualified": yes_no,
        "ee_qualified": yes_no,
    },
    "table-canadarelative": {
        "province": canada_provinces,
        "status": imm_status,
        "relationship": relationship,
        "sex": sex,
        "age": positive_int,
        "years_in_canada": positive_int,
    },
    "table-family": {
        "marital_status": marital_status,
        "date_of_birth": date_format,
        "relationship": family_relationship,
        "date_of_death": date_format,
        "accompany_to_canada": yes_no,
    },
    "table-cor": {
        "start_date": date_format,
        "end_date": date_format,
        "status": imm_status,
    },
    "table-travel": {
        "start_date": date_format,
        "end_date": date_format,
        "length": positive_int,
    },
    "table-military": {"start_date": date_format, "end_date": date_format},
    "table-illtreatment": {"start_date": date_format, "end_date": date_format},
    "table-member": {"start_date": date_format, "end_date": date_format},
    "table-government": {"start_date": date_format, "end_date": date_format},
    # PR
    "info-prcase": {
        "submission_date": date_format,
        "last_entry_date": date_format,
        "imm_program": pr_imm_program,
        "imm_category": pr_imm_category,
        "imm_under": pr_imm_under,
        "communication_language": english_or_french,
        "interview_language": english_french_chinese,
        "need_translator": yes_no,
        "intended_province": canada_provinces,
        "has_csq": yes_no,
        "consent_of_info_release": yes_no,
        "number": positive_int,
    },
    "info-background": {
        "q1a": yes_no,
        "q1b": yes_no,
        "q1c": yes_no,
        "q2a": yes_no,
        "q2b": yes_no,
        "q2c": yes_no,
        "q3a": yes_no,
        "q4a": yes_no,
        "q5": yes_no,
        "q6": yes_no,
    },
    # Recruitment
    "table-advertisement": {"start_date": date_format, "end_date": date_format},
    "table-interviewrecord": {
        "canadian_status": interview_canadian_status,
        "interviewed": yes_no,
        "offered": yes_no,
        "accepted": yes_no,
    },
    "info-recruitmentsummary": {
        "reply2apply": yes_no,
        "emails_for_making_interview": yes_no,
        "interview_record": yes_no,
        "interview_process_evidence": yes_no,
        "emails_for_certificates": yes_no,
        "emais_for_references": yes_no,
        "reference_checked": yes_no,
        "reference_check_evidence": yes_no,
        "joboffer_email": yes_no,
        "joboffer_email_reply": yes_no,
        "after_offer_coomunication": yes_no,
        "interview_date": date_format,
    },
    # TR
    "info-trcasein": {
        "service_in": english_french,
        "original_entry_date": date_format,
        "most_recent_entry_date": date_format,
        "original_purpose": tr_original_purpose,
        "is_spouse_canadian": yes_no,
        "consent_of_info_release": yes_no,
        "submission_date": date_format,
    },
    "info-trcase": {
        "service_in": english_french,
        "application_purpose": tr_application_purpose,
        "same_as_cor": yes_no,
        "applying_stauts": imm_status,
        "applying_start_date": date_format,
        "applying_end_date": date_format,
        "submission_date": date_format,
    },
    "info-sp": {
        "study_level": education_level,
        "province": canada_provinces,
        "paid_person": sp_paid_person,
        "duration_from": date_format,
        "duration_to": date_format,
        "tuition_cost": positive_int_decimal,
        "room_cost": positive_int_decimal,
        "other_cost": positive_int_decimal,
        "fund_available": positive_int_decimal,
    },
    "info-vrincanada": {
        "application_purpose": vr_application_purpose,
        "duration_from": date_format,
        "duration_to": date_format,
        "funds_available": positive_int_decimal,
    },
    "info-spincanada": {
        "application_purpose": sp_in_application_purpose,
        "study_level": education_level,
        "province": canada_provinces,
        "paid_person": sp_paid_person,
        "apply_work_permit": yes_no,
        "work_permit_type": sp_apply_wp_type,
        "duration_from": date_format,
        "duration_to": date_format,
        "tuition_cost": positive_int_decimal,
        "room_cost": positive_int_decimal,
        "other_cost": positive_int_decimal,
        "fund_available": positive_int_decimal,
        "expiry_date": date_format,
    },
    "info-wpincanada": {
        "application_purpose": wp_in_application_purpose,
        "work_province": canada_provinces,
        "work_permit_type": wp_apply_wp_type,
        "duration_from": date_format,
        "duration_to": date_format,
        "pnp_certificated": yes_no,
        "expiry_date": date_format,
    },
    "info-wp": {
        "work_province": canada_provinces,
        "work_permit_type": wp_apply_wp_type,
        "duration_from": date_format,
        "duration_to": date_format,
    },
    "info-visa": {
        "application_purpose": visa_application_purpose,
        "duration_from": date_format,
        "duration_to": date_format,
        "funds_available": positive_int_decimal,
    },
    "info-incanadacommon": {
        "original_purpose": tr_original_purpose,
        "original_entry_date": date_format,
        "most_recent_entry_date": date_format,
    },
    "info-trbackground": {
        "q1a": yes_no,
        "q1b": yes_no,
        "q2a": yes_no,
        "q2b": yes_no,
        "q2c": yes_no,
        "q3a": yes_no,
        "q4a": yes_no,
        "q5": yes_no,
        "q6": yes_no,
    },
    # PR
    "info-prbackground": {
        "q1": yes_no,
        "q2": yes_no,
        "q3": yes_no,
        "q4": yes_no,
        "q5": yes_no,
        "q6": yes_no,
        "q7": yes_no,
        "q8": yes_no,
        "q9": yes_no,
        "q10": yes_no,
        "q11": yes_no,
    },
}



from datetime import date
from pydantic import BaseModel, EmailStr
from typing import List, Union


class COR(BaseModel):
    start_date: date
    end_date: Union[date, None]
    country: str
    status: str


class CORs(object):
    # Frist row must be current residence
    def __init__(self, cors: List[COR]):
        self.cors = cors

    @property
    def current(self):
        ccor = [country for country in self.cors if country.end_date == None]
        if len(ccor) == 1:
            return ccor[0]
        elif len(ccor) == 0 and len(self.cors) > 0:
            return self.cors[0]
        else:
            raise ValueError("No residence data, please check")

    @property
    def previous(self):
        if len(self.cors) == 0:
            print("There is no residence data, please check")
            return
        return self.cors[1:]



from datetime import datetime,date
from enum import Enum, unique
from difflib import Differ
import re
from fuzzywuzzy import fuzz
from termcolor import colored

# This function used to check a list's continuity of date. The list must have two columns as start date and end date.
# The fromat must be "yyyy-mm-dd" or "yyy-mm".  The list has to be from most recent to past. You can force sort it by assign sort =True
# The output will be True or False, plus sorted input list and message.
def checkContinuity(data_set: list, sort=False):
    if sort:
        data_set = sorted(data_set, key=lambda x: (x[0], x[1]), reverse=True)
    ok = []
    errors = []

    def difference(l1, l2):
        if type(l1)==date: l1=l1.strftime("%Y-%m-%d")
        if type(l2)==date: l2=l2.strftime("%Y-%m-%d")

        if l2==None or l2=="Present":
            l2=datetime.today().strftime("%Y-%m-%d") if bool(re.search(r"^\d{4}-\d{2}-\d{2}", l1)) else datetime.today().strftime("%Y-%m")
        if bool(re.search(r"^\d{4}-\d{2}-\d{2}", l1)) and bool(
            re.search(r"^\d{4}-\d{2}-\d{2}", l2)
        ):
            current_start = datetime.strptime(l1, "%Y-%m-%d")
            previous_end = datetime.strptime(l2, "%Y-%m-%d") 
            return (current_start - previous_end).days, "day(s)"
        if bool(re.search(r"^\d{4}-\d{2}", l1)) and bool(
            re.search(r"^\d{4}-\d{2}", l2)
        ):
            y1, m1 = l1.split("-")
            y2, m2 = l2.split("-")
            return (int(y1) - int(y2)) * 12 + int(m1) - int(m2), "month(s)"

    for i in range(len(data_set) - 1):
        diff, unit = difference(data_set[i][0], data_set[i + 1][1])
        if diff > 1:
            ok.append(False)
            errors.append(
                f"{'List sorted. ' if sort else ''}The date is not continious between line {i+1} to {i+2} (missed {diff-1} {unit})"
            )
        elif diff < 0:
            ok.append(False)
            errors.append(
                f"{'List sorted. ' if sort else ''}The date is not continious between line {i+1} to {i+2} (overlapped {diff} {unit})"
            )
        else:
            ok.append(True)
    if all(ok):
        return True, data_set, "date is continious"
    else:
        return False, data_set, errors


#  Used to find different between two text strings. Initialize with tc=TextChanged(t1,t2), and get differences by tc.text_added or tc.text_deleted
class TextChanged:
    def __init__(self, text1, text2):
        self.text1 = text1
        self.text2 = text2

    def __get_changed(self):

        t1_lines = self.text1.splitlines()
        t2_lines = self.text2.splitlines()

        diff = Differ().compare(t1_lines, t2_lines)
        changed = list(diff)

        added = r"^\+\s(.+)"
        deleted = r"^\-\s(.+)"
        text_added = [
            d[2:] for d in changed if re.findall(added, d)
        ]  # d[2:] delete '+ ' at the begining
        text_added = "\n".join(text_added)
        text_deleted = [d[2:] for d in changed if re.findall(deleted, d)]
        text_deleted = "\n".join(text_deleted)

        return {"text_added": text_added, "text_deleted": text_deleted}

    @property
    def text_added(self):
        return self.__get_changed()["text_added"]

    @property
    def text_deleted(self):
        return self.__get_changed()["text_deleted"]


def best_match(term, list_names, min_score=0):
    max_score = -1
    max_name = ""
    for term2 in list_names:
        score = fuzz.ratio(term, term2)
        if (score > min_score) and (score > max_score):
            max_name = term2
            max_score = score
    return max_name


Month = Enum(
    "Month",
    (
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ),
)


@unique
class Sex(Enum):
    Female = 0
    Male = 1


# input list and outpout markdown table text
def markdown_table(ds: list):
    head = ds[0]
    head_str = "|"
    head_sep = "|"
    for col in range(0, len(ds[0])):
        head_str += str(head[col]) + "|"
        head_sep += "------------- " + "|"
    # print(head_str)
    # print(head_sep)
    output = head_str + "\n" + head_sep + "\n"

    row_data = "|"
    for row in range(1, len(ds)):
        for col in range(0, len(ds[0])):
            row_data += str(ds[row][col]) + "|"
        # print(row_data)
        output += row_data + "\n"
        row_data = "|"

    return output


# Some common input functions
def multi_line_input():
    print("Ctrl-D  to save it.")
    contents = []
    while True:
        try:
            line = input()
        except EOFError:
            break
        contents.append(line)
    return "\n".join(contents)




skill_levels = {
    '0': 'A',  # if the second number is 0 or 1, the level is A
    '1': 'A',
    '2': 'B',
    '3': 'B',
    '4': 'C',
    '5': 'C',
    '6': 'D',
    '7': 'D'
}

def getNOCLevel(noc):
    level = ''
    if noc[0] == '0':
        if noc[1] == '0':
            level = '00'
        else:
            level = '0'
    else:
        level = skill_levels.get(noc[1],None)
    return level

# key is factor, value is points. Most common scoring way
def getKeyValuePoint(points_dict,factor_key):
    return points_dict.get(factor_key,0)

# points range is a 2-dimensions list. 1st column and 2nd column are range, the 3rd column is value
def getRangePoint(points_range,factor):
    for i in range(len(points_range)):
        if factor in range(points_range[i][0], points_range[i][1]):
            return (points_range[i][2])
    return 0


# points is a dict or a list including all points, and this function returns the all points sum 
def getTotalPoint(points):
    return_point=0
    if isinstance(points,dict):
        for point in points.values():
            return_point+=point
    elif isinstance(points,list):
        for point in points:
            return_point+=point
    else:
        return_point=0
    return return_point


from dataclasses import dataclass
from datetime import date
from email.errors import ObsoleteHeaderDefect
from pydantic import BaseModel
from typing import Optional, List
from model.common.address import Address


class TrCase(BaseModel):
    service_in: str
    application_purpose: Optional[str]  # 5257需要，1294 5 不需要
    same_as_cor: bool
    applying_country: Optional[str]
    applying_status: Optional[str]
    applying_start_date: Optional[date]
    applying_end_date: Optional[date]
    consent_of_info_release: bool
    submission_date: Optional[date]


class TrCaseIn(BaseModel):
    service_in: str
    application_purpose: Optional[str]  # TODO: consider
    original_entry_date: date
    original_entry_place: str
    original_purpose: str
    original_other_reason: Optional[str]
    most_recent_entry_date: date
    most_recent_entry_place: str
    doc_number: Optional[str]
    consent_of_info_release: bool
    submission_date: Optional[date]


class Visa(BaseModel):
    visit_purpose: str
    duration_from: date
    duration_to: date
    funds_available: int
    name1: Optional[str]
    relationship1: Optional[str]
    address1: Optional[str]
    name2: Optional[str]
    relationship: Optional[str]
    address2: Optional[str]


class Sp(BaseModel):
    school_name: str
    study_level: str
    study_field: str
    province: str
    city: str
    address: str
    dli: str
    student_id: str
    duration_from: date
    duration_to: date
    tuition_cost: Optional[str]
    room_cost: Optional[str]
    other_cost: Optional[str]
    fund_available: str
    paid_person: str
    other: Optional[str]


class Wp(BaseModel):
    work_permit_type: str
    other_explain: Optional[str]
    employer_name: Optional[str]
    employer_address: Optional[str]
    work_province: Optional[str]
    work_city: Optional[str]
    work_address: Optional[str]
    job_title: Optional[str]
    brief_duties: Optional[str]
    duration_from: date
    duration_to: date
    lmia_num_or_offer_num: Optional[str]
    # pnp_certificated: bool
    caq_number: Optional[str]
    expiry_date: Optional[date]


class VrInCanada(BaseModel):
    application_purpose: str
    visit_purpose: str
    other_explain: Optional[str]
    duration_from: date
    duration_to: date
    funds_available: int
    paid_by: str
    other_payer_explain: Optional[str]
    name1: Optional[str]
    relationship1: Optional[str]
    address1: Optional[str]
    name2: Optional[str]
    relationship2: Optional[str]
    address2: Optional[str]


class SpInCanada(BaseModel):
    application_purpose: str
    apply_work_permit: bool
    work_permit_type: Optional[str]
    caq_number: Optional[str]
    expiry_date: Optional[date]
    school_name: str
    study_level: str
    study_field: str
    province: str
    city: str
    address: str
    dli: str
    student_id: str
    duration_from: date
    duration_to: date
    tuition_cost: Optional[str]
    room_cost: Optional[str]
    other_cost: Optional[str]
    fund_available: str
    paid_person: str
    other: Optional[str]


class WpInCanada(BaseModel):
    application_purpose: str
    caq_number: Optional[str]
    expiry_date: Optional[date]
    work_permit_type: str
    employer_name: Optional[str]
    employer_address: Optional[str]
    work_province: Optional[str]
    work_city: Optional[str]
    work_address: Optional[str]
    job_title: Optional[str]
    brief_duties: Optional[str]
    duration_from: Optional[date]
    duration_to: Optional[date]
    lmia_num_or_offer_num: Optional[str]
    pnp_certificated: bool


class TrBackground(BaseModel):
    q1a: bool
    q1b: bool
    q1c: Optional[str]
    q2a: bool
    q2b: bool
    q2c: bool
    q2d: Optional[str]
    q3a: bool
    q3b: Optional[str]
    q4a: bool
    q4b: Optional[str]
    q5: bool
    q6: bool



# Engine class can be used to fill all web pages, as long as the import model matches our model prototype, and already contains value. 
# Every page has to make its own page data tree model , and assembly values together.  So, the only import data is assemblied model and data sets. 
# The class takes charge of identifing element type and do related action (click, select, and finally fill text fields...)
from operator import contains
from models.definition import Action
from models.element import (RadioYesNoElement,RadioYesNoNullElement,ButtonElement,
                            CheckboxElement,SelectElement, InputElement,AreatextElement,
                            TurnPage,DependantSelectElement)

class Engine():
    def __init__(self):
        self.output = [] # js app handleable objects list
    
    # recurssion for manupilating all elements. It's common for all pages, so put these functions into the Page prototype
    def treeForeach(self,model):
        for node in model:
            self.handleNode(node)
            children = self.getChildren(node)
            children and self.treeForeach(children) # 遍历子树

    # This function focuses on processing node founded by getChildren
    def handleNode(self,node):
        match node.get("action_type"):
            case Action.Radio_Yes_No.value:
                # 根据模型和数据，生成puppeteer能处理的对象
                radio_yn=RadioYesNoElement(**node)
                self.output.append(radio_yn.make())
                
            case Action.Radio_Yes_No_Null.value:
                radio_ynn=RadioYesNoNullElement(**node)
                self.output.append(radio_ynn.make())
                    
            case Action.Checkbox.value:
                cbk=CheckboxElement(**node)
                self.output.append(cbk.make())
            
            case Action.Select.value:
                select=SelectElement(**node)
                self.output.append(select.make())
            
            case Action.SelectPopup.value: 
                select=SelectElement(**node)
                self.output.append(select.make())
                
            case Action.Input.value:
                input=InputElement(**node)
                self.output.append(input.make())  
                
            case Action.Areatext.value:
                areatext=AreatextElement(**node)
                self.output.append(areatext.make())  
            
            case Action.Button.value:
                button=ButtonElement(**node)
                self.output.append(button.make()) 
                
            case Action.Turnpage.value:
                turnpage=TurnPage(**node)
                self.output.append(turnpage.make())  
                
            case Action.RepeatSection.value:
                self.repeatSection(node)
            
            case Action.DependantSelect.value:
                ds=DependantSelectElement(**node)
                self.output.append(ds.make()) 
            
            case Action.SelectPopup2.value: # dependanat select and followed by children
                #1 get dependant select element
                dse={
                    "action_type":"DependantSelectElement",
                    "select1":node.get('select1'),
                    "select2":node.get('select2')
                }
                select2=DependantSelectElement(**dse)
                self.output.append(select2.make())  

            case _:
                pass

    # This funciton only takes charge of finding nodes, not processing it.
    def getChildren(self,node):
        match node.get("action_type"):
            case Action.Radio_Yes_No.value:
                return node.get("yes_children") if node.get("value") else node.get("no_children")
            case Action.Radio_Yes_No_Null.value:
                if node.get("value") != None:
                    return node.get("yes_children") if node.get("value") else node.get("no_children")
                return node.get("null_children")
            case Action.SelectPopup.value:
                #according to the selection, choose  children
                childName = '_'.join(node.get("nameValue").split('-')).lower() + "_children" if "-" in node.get("nameValue")  else "_".join(node.get('nameValue').split(' ')).lower() + "_children"
                return node[childName]
            case Action.SelectPopup2.value:
                #according to the selection, choose  children
                return node[node.get('generated_children_name')]
            case _:
                return node.get("children")
    
    # handle repeat sections
    # 限制：repeat内部的元素，必须是最底层element,不能再有下级元素。
    # valueObjects是客户数据，targetElements是网页元素，本程序匹配他们生成最终填表列表数据
    def repeatSection(self,node):
        # loop the valueObjects, and create new keys based on targetElements, then meke new list of objects by assigning value to new keys
        for index, value_object in enumerate(node.get("valueObjects")):
            for target_element in node.get("targetElements"):
                # 从target emement的value处，获得value object的key
                key=target_element['value']
                #element.value is the key of value object, so get the final value for web element
                value=value_object[key]  
                # targetElements中, ???是通配符，需要用index来替代。
                newElementId = target_element['id'].replace("???",str(index))
                # 将获得了新ID和value的元素替代老的
                newElement = {**target_element, **{ "id": newElementId, "value": value } }
                self.handleNode(newElement);


    


from typing import List, Optional, Union
from model.common.address import Address
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from datetime import date
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country_residence import country_residence

class General(BaseModel):
    legal_name:str
    operating_name:Optional[str]
    website:Optional[str]
    recruit_email:EmailStr
    company_intro:str
    business_intro:str


class LMIAPrModel(CommonModel):
    general:General
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels=self.getExcels(['excel/er.xlsx','excel/pa.xlsx'])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())

class LMIAHWSModel(CommonModel):
    pass



from model.tr.d1294 import data_model, convert_model,remove_model
from model.tr.pdfformmodel import PdfFormModel


class F1294Model(PdfFormModel):
    
    def __init__(self,xml_file):
        # pass the globals variables to base class
        model={'data_model':data_model,'convert_model':convert_model,'remove_model':remove_model}
        super().__init__(xml_file,model)

    def _getSpecial(self):
        # work
        occupation_info = [
            e for e in self.tree.iterfind('Page3/Occupation')][0]
        occupations = []
        for i in [1, 2, 3]:
            occupation_item = occupation_info.find('OccupationRow'+str(i))
            occupation = self._getSubItem(
                occupation_item, data_model['occupation'])
            occupations.append(occupation)

        self.data['occupations'] = occupations





from functools import reduce
from pydantic import BaseModel, ValidationError
from termcolor import colored
from source.excel import Excel
from model.common.errordict import error_dict
from .context import DATADIR


class CommonModel(BaseModel):
    # every succesor class has its own properties
    class Config:
        arbitrary_types_allowed = True

    # initialize the model with a list of excels, which includes all nececcery information the model required.
    # output_excel_file is finename for generating new excel. global_dict is the dict generated by globals() from successor's class
    def __init__(self, excels, output_excel_file, global_dict, generate_excel=True):
        # add up all input excels and return the combined one
        *original_excel_objs, new_excel_obj = self._addupExcels(
            excels
        )  # the last one is new_excel_obj, before this are all original_objs
        if output_excel_file:
            data = self._generateExcel(global_dict, new_excel_obj, output_excel_file)
            if generate_excel:
                self.generateExcel(**data)
        else:
            # get the dict as input kwargs. Here globals() is very important. Without it, the KwargsMix cannot find required classes
            kwargs = self._getInput(global_dict, original_excel_objs, new_excel_obj)
            # use pre-assigned referrence value (kwargs) to initialize the model
            try:
                super().__init__(**kwargs)
            except ValidationError as e:
                print(e)
                exit(1)

    # get combined excel throgh add up all input excels.
    def _addupExcels(self, excels):
        excel_objs = list(map(lambda x: Excel(x), excels))
        return [*excel_objs, reduce(lambda a, b: a + b, excel_objs)]

    """
    Prerequisite of below methods
    1. excel_obj is the one that all input excels combined, no sheet has same name
    2. Successor's model is constructed with all sheets/tables, no specific single varialbe
    3. Only InfoSheet and TableList two data structure in the successor's model properties
    """

    def _getInput(self, global_dict, original_excel_objs, new_excel_obj):

        input_kwargs = {}
        # According to the prerequisite described above, all the referrenced scheme are defined in 'definitions'.
        # key is Model name, and after lower(), it is also excel sheet name and the property defined in this class. They are same. This is the requirement.
        for k in self.schema()["definitions"].keys():
            # globals() returns a dict inlcuding all global classes, methods, ....
            klass = global_dict.get(k)
            # use k.lower (it's a sheet name here) to get sheet data
            sheet_data = new_excel_obj.dict.get(k.lower(), None)
            # only two types, one is list of model objects,another is model object
            if not isinstance(sheet_data, list):
                # assign the input kwargs. Here k.lower() is class's property. klass is the model, on which the app create an obj
                try:
                    input_kwargs[k.lower()] = klass(**sheet_data)
                except ValidationError as e:
                    sheet_name = "info-" + k.lower()  # sheet name
                    excel_obj = [
                        excel_obj
                        for excel_obj in original_excel_objs
                        if sheet_name in excel_obj.sheet_names
                    ][0]
                    self.interpretError(excel_obj, sheet_name, e)
            else:
                try:
                    input_kwargs[k.lower()] = [klass(**d) for d in sheet_data]
                except ValidationError as e:
                    table_name = "table-" + k.lower()  # sheet name
                    excel_obj = [
                        excel_obj
                        for excel_obj in original_excel_objs
                        if table_name in excel_obj.table_names
                    ][0]
                    self.interpretError(excel_obj, table_name, e)
        return input_kwargs

    # generate a new excel according to the program class's properties
    def _generateExcel(self, global_dict, new_excel_obj, excel_file):
        sheets = []
        tables = []
        # successo's referrenced class is retored in schema's definition fields with dict format.
        # 获得每一sheet对应class，并获得它的properties list
        classes = self.schema()["definitions"].keys()
        for k in classes:
            # global_dict is a dict inlcuding all global classes, methods, ....
            klass = global_dict.get(
                k
            )  # k here is actually a sheet/table name, and klass is the class related to sheet.
            variables = list(
                klass.__fields__.keys()
            )  # get all variables in the sheet/table
            sheet_table_name = (
                k.lower()
            )  # k.lower is the sheet / table name according to convention
            if "info-" + sheet_table_name in new_excel_obj.sheet_names:
                sheets.append(
                    new_excel_obj.getSheet(sheet_table_name, variables)
                )  # get sheet object
            if "table-" + sheet_table_name in new_excel_obj.table_names:
                tables.append(
                    new_excel_obj.getTable(sheet_table_name, variables)
                )  # get table object
        return {
            "new_excel_obj": new_excel_obj,
            "excel_file": excel_file,
            "sheets": sheets,
            "tables": tables,
        }

    def generateExcel(self, **data):
        # 根据最终更新了variables的sheets和talbes，生成新的excel文件
        data["new_excel_obj"].makeExcel(
            data["excel_file"], sheets=data["sheets"], tables=data["tables"]
        )

    # def interpretError(self,excel_obj,sheet_name,e:ValidationError):
    #     excel_name=excel_obj.excel_name.split('/')[-1] # only get the excel file name and ignore the path
    #     print(colored(f"{excel_name}{error_dict.get('file')}{sheet_name}{error_dict.get('sheet')}{error_dict.get('has error')}，{error_dict.get('detail')}: ",'red'))
    #     for err in e.errors():
    #         for variable in err['loc']:
    #             if sheet_name.startswith('info-'):
    #                 sheet=excel_obj.getSheet(sheet_name[5:]) # remove info-
    #                 info_node=sheet.data.get(variable)
    #                 if info_node !=None:
    #                     print(colored(f"{info_node.description}",'red'),error_dict.get(err['msg']),error_dict.get('has error'))
    #             if sheet_name.startswith('table-'):
    #                 table=excel_obj.getTable(sheet_name[6:])
    #                 title=table.column_titles.get(variable)
    #                 print(colored(f"{title}",'red'),error_dict.get('the_column'),error_dict.get('has error'),error_dict.get(err['msg'],error_dict.get('error')))
    #     print(e.json)

    def interpretError(self, excel_obj, sheet_name, e: ValidationError):
        excel_name = excel_obj.excel_name.split("/")[
            -1
        ]  # only get the excel file name and ignore the path
        print(
            colored(
                f"{excel_name}{error_dict.get('file')}{sheet_name}{error_dict.get('sheet')}{error_dict.get('has error')}，{error_dict.get('detail')}: ",
                "red",
            )
        )
        for err in e.errors():
            for variable in err["loc"]:
                if sheet_name.startswith("info-"):
                    sheet = excel_obj.getSheet(sheet_name[5:])  # remove info-
                    info_node = sheet.data.get(variable)
                    if info_node != None:
                        print(colored(f"{info_node.description}", "red"), err["msg"])
                if sheet_name.startswith("table-"):
                    table = excel_obj.getTable(sheet_name[6:])
                    title = table.column_titles.get(variable)
                    print(
                        colored(f"{title}", "red"), "the_column has error", err["msg"]
                    )
        print(e)

    def getExcels(self, templates=[]):
        excels = []
        import os

        for template in templates:
            excel_file = os.path.abspath(os.path.join(DATADIR, template))
            excels.append(excel_file)
        return excels



from webform.bcpnp.dashboard import DashboardReg
from webform.models.definition import Action
from webform.models.bcpnpmodel_reg import BcpnpModelReg, BcpnpEEModelReg
from model.common.address import Addresses
from webform.bcpnp.data import city_map
from typing import Union


class Joboffer:
    def __init__(self, person: Union[BcpnpModelReg, BcpnpEEModelReg]):
        self.person = person

    def fill(self):
        company_details = [
            {
                "action_type": Action.Input.value,
                "label": "Company legal name",
                "id": "#syncA_App_Emp_Comp_LegalName",
                "value": self.person.general.legal_name,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Company operating name",
                "id": "#syncA_App_Emp_Comp_OperName",
                "value": self.person.general.operating_name,
                "required": False,
                "length": 100,
            },
        ]
        address = Addresses(self.person.eraddress).working
        work_location = [
            {
                "action_type": Action.Input.value,
                "label": "Unit number",
                "id": "#BCPNP_App_Job_WorkLocationAddrUnit",
                "value": address.unit,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Street address",
                "id": "#BCPNP_App_Job_WorkLocationAddr",
                "value": address.street_address,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Select.value,
                "label": "City/Town",
                "id": "#syncA_App_Job_WorkLocationCity",
                "value": city_map[address.city],
            },
            {
                "action_type": Action.Input.value,
                "label": "Postal code",
                "id": "#syncA_App_Job_WorkLocationPostal",
                "value": address.post_code,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Phone number",
                "id": "#BCPNP_App_Job_WorkLocationPhone",
                "value": self.person.joboffer.phone,
                "required": True,
                "length": 100,
            },
        ]

        joboffer_details = [
            {
                "action_type": Action.Input.value,
                "label": "Job title",
                "id": "#syncA_App_Job_Title",
                "value": self.person.joboffer.job_title,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "NOC",
                "id": "#syncA_App_Job_NOC",
                "value": self.person.joboffer.noc,
                "required": False,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Hours of work per week",
                "id": "#syncA_App_Job_HoursPerWeek",
                "value": self.person.joboffer.weekly_hours,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Hourly wage",
                "id": "#syncA_App_Job_HourlyWage",
                "value": self.person.joboffer.hourly_rate,
                "required": False,
                "length": 100,
            },
            {
                "action_type": Action.Input.value,
                "label": "Annual Salary",
                "id": "#syncA_App_Job_AnnualWage",
                "value": self.person.joboffer.annual_rate,
                "required": True,
                "length": 100,
            },
            {
                "action_type": Action.Radio.value,
                "label": "Are you currently working for the employer in the job being offered? ",
                "id": "#BCPNP_App_Job_Work_AtOfferedJob-Yes"
                if self.person.joboffer.is_working
                else "#BCPNP_App_Job_Work_AtOfferedJob-No",
            },
        ]

        working_fulltime = (
            [
                {
                    "action_type": Action.Radio.value,
                    "label": "Are you working full-time in B.C. in the job being offered? ",
                    "id": "#BCPNP_App_Job_Work_CurrentOffer-Yes"
                    if float(self.person.joboffer.weekly_hours) >= 30
                    else "#BCPNP_App_Job_Work_CurrentOffer-No",
                },
            ]
            if self.person.joboffer.is_working
            else []
        )

        # dashboard = DashboardReg()
        actions = (
            # dashboard.jump("Job Offer")
            company_details
            + work_location
            + joboffer_details
            + working_fulltime
            # + dashboard.save
        )
        return [
            {
                "action_type": Action.WebPage.value,
                "page_name": "Joboffer",
                "actions": actions,
                "id": "body > div > main > div.layout-container > div > div > div > uf-form > div > ng-transclude > div > uf-tabset > div > ul > li:nth-child(5) > a",
                "next_page_tag":"#BCPNP_App_LangTest_Completed-No"
            }
        ]



from functools import reduce
from pydantic import BaseModel, validator
from datetime import date
from typing import Optional, List
from model.common.mixins import DurationMixin
from model.common.utils import makeList, normalize

"""
with properties of start_from and end_to got from mixin
"""
education_level = {
    "None": 0,
    "High school": 1,
    "Certificate/Diploma(Trade)": 2,
    "Diploma/Certificate": 3,
    "Associate": 4,
    "Bachelor": 5,
    "Post-graduate diploma": 6,
    "Master": 7,
    "Doctor": 8,
}


class EducationBase(BaseModel, DurationMixin):
    start_date: date
    end_date: Optional[date]
    school_name: str
    education_level: str
    field_of_study: str

    _normalize_school_name = validator(
        "school_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_field_of_study = validator(
        "field_of_study", allow_reuse=True, check_fields=False
    )(normalize)

    @validator("end_date")
    def endDateBigger(cls, end_date, values):
        start_date = values.get("start_date")
        the_date = end_date if end_date else date.today()
        if (the_date - start_date).days <= 0:
            raise ValueError(
                f"End date {the_date} is earlier than start date {start_date}"
            )
        end_date = end_date if end_date else "Present"
        return end_date

    @property
    def level_in_num(self):
        return education_level[self.education_level]


class EducationHistory:
    def __init__(self, edu_list: List[EducationBase]):
        self.edu_list = edu_list

    # total education years
    @property
    def years(self):
        return reduce(lambda a, b: a.lengthOfYears + b.lengthOfYears, self.edu_list)

    @property
    def highest(self):

        highest_num = max(
            [education_level[edu.education_level] for edu in self.edu_list]
        )
        for k, v in education_level.items():
            if v == highest_num:
                return k

    @property
    def highestEducation(self):
        highest_edu = [
            edu
            for edu in self.edu_list
            if edu.education_level == self.highest and edu.end_date
        ]  # edu.end_date should not be empty, which means it's ongoing education.
        if len(highest_edu) > 0:
            return highest_edu[0]

    # return highest post-secondary education only
    @property
    def highestPostSecondaryEducation(self):
        highest_edu = [
            edu
            for edu in self.edu_list
            if edu.education_level == self.highest
            and edu.end_date
            and education_level[edu.education_level] >= 2
        ]
        if len(highest_edu) > 0:
            return highest_edu[0]

    @property
    def high_school(self):
        return [edu for edu in self.edu_list if edu.education_level == "High school"]

    @property
    def post_secondary(self):
        return [
            edu
            for edu in self.edu_list
            if edu.education_level not in ["High school", "None"]
        ]

    def post_secondary_in_ca_prov(self, province):
        return [
            edu
            for edu in self.edu_list
            if edu.education_level not in ["High school", "None"]
            and edu.province == province
        ]

    def post_secondary_in_ca_but_not_in_prov(self, province):
        return [
            edu
            for edu in self.edu_list
            if edu.education_level not in ["High school", "None"]
            and edu.province != province
            and edu.country == "Canada"
        ]

    @property
    def post_secondary_not_in_ca(self):
        return [
            edu
            for edu in self.edu_list
            if edu.education_level not in ["High school", "None"]
            and edu.country != "Canada"
        ]
